@@ -2,6 +2,7 @@
 
 # Copyright (c) 2009-2014 - Simon Conseil
 # Copyright (c) 2013      - Christophe-Marie Duquesne
+# Copyright (c) 2014      - Jonas Kaufmann
 
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to
@@ -27,21 +28,23 @@ import fnmatch
 import logging
 import multiprocessing
 import os
+import pickle
 import sys
 import zipfile
 
+from click import progressbar, get_terminal_size
 from collections import defaultdict
 from datetime import datetime
-from functools import partial
+from itertools import cycle
 from os.path import isfile, join, splitext
 from PIL import Image as PILImage
 
 from . import image, video, signals
-from .compat import UnicodeMixin, strxfrm, url_quote
-from .image import process_image, get_exif_tags
-from .log import colored, BLUE
+from .compat import UnicodeMixin, strxfrm, url_quote, text_type
+from .image import process_image, get_exif_tags, get_exif_data
 from .settings import get_thumb
-from .utils import copy, check_or_create_dir, url_from_path, read_markdown
+from .utils import (Devnull, copy, check_or_create_dir, url_from_path,
+                    read_markdown, cached_property)
 from .video import process_video
 from .writer import Writer
 
@@ -68,6 +71,7 @@ class Media(UnicodeMixin):
         self.src_filename = self.filename = self.url = filename
         self.path = path
         self.settings = settings
+        self.ext = os.path.splitext(filename)[1].lower()
 
         self.src_path = join(settings['source'], path, filename)
         self.dst_path = join(settings['destination'], path, filename)
@@ -76,9 +80,6 @@ class Media(UnicodeMixin):
         self.thumb_path = join(settings['destination'], path, self.thumb_name)
 
         self.logger = logging.getLogger(__name__)
-        self.raw_exif = None
-        self.exif = None
-        self.date = None
         self._get_metadata()
         signals.media_initialized.send(self)
 
@@ -91,10 +92,18 @@ class Media(UnicodeMixin):
     @property
     def big(self):
         """Path to the original image, if ``keep_orig`` is set (relative to the
-        album directory).
+        album directory). Copy the file if needed.
         """
         if self.settings['keep_orig']:
-            return os.path.join(self.settings['orig_dir'], self.src_filename)
+            s = self.settings
+            if s['use_orig']:
+                # The image *is* the original, just use it
+                return self.filename
+            orig_path = join(s['destination'], self.path, s['orig_dir'])
+            check_or_create_dir(orig_path)
+            copy(self.src_path, join(orig_path, self.src_filename),
+                 symlink=s['orig_link'])
+            return url_from_path(join(s['orig_dir'], self.src_filename))
         else:
             return None
 
@@ -138,11 +147,24 @@ class Image(Media):
     type = 'image'
     extensions = ('.jpg', '.jpeg', '.png')
 
-    def __init__(self, filename, path, settings):
-        super(Image, self).__init__(filename, path, settings)
-        self.raw_exif, self.exif = get_exif_tags(self.src_path)
-        if self.exif is not None and 'dateobj' in self.exif:
-            self.date = self.exif['dateobj']
+    @cached_property
+    def date(self):
+        return self.exif and self.exif.get('dateobj', None) or None
+
+    @cached_property
+    def exif(self):
+        return (get_exif_tags(self.raw_exif)
+                if self.raw_exif and self.ext in ('.jpg', '.jpeg') else None)
+
+    @cached_property
+    def raw_exif(self):
+        try:
+            return (get_exif_data(self.src_path)
+                    if self.ext in ('.jpg', '.jpeg') else None)
+        except (IOError, IndexError, TypeError, AttributeError):
+            self.logger.warning(u'Could not read EXIF data from %s',
+                                self.src_path)
+            return None
 
 
 class Video(Media):
@@ -154,6 +176,7 @@ class Video(Media):
     def __init__(self, filename, path, settings):
         super(Video, self).__init__(filename, path, settings)
         base = splitext(filename)[0]
+        self.date = None
         self.src_filename = filename
         self.filename = self.url = base + '.webm'
         self.dst_path = join(settings['destination'], path, base + '.webm')
@@ -177,14 +200,13 @@ class Album(UnicodeMixin):
     """
 
     description_file = "index.md"
-    output_file = 'index.html'
 
     def __init__(self, path, settings, dirnames, filenames, gallery):
         self.path = path
         self.name = path.split(os.path.sep)[-1]
         self.gallery = gallery
         self.settings = settings
-        self.orig_path = None
+        self.output_file = settings['output_filename']
         self._thumbnail = None
 
         if path == '.':
@@ -224,16 +246,6 @@ class Album(UnicodeMixin):
             self.medias_count[media.type] += 1
             medias.append(media)
 
-        # sort images
-        if medias:
-            medias_sort_attr = settings['medias_sort_attr']
-            if medias_sort_attr == 'date':
-                key = lambda s: s.date or datetime.now()
-            else:
-                key = lambda s: strxfrm(getattr(s, medias_sort_attr))
-
-            medias.sort(key=key, reverse=settings['medias_sort_reverse'])
-
         signals.album_initialized.send(self)
 
     def __repr__(self):
@@ -281,6 +293,18 @@ class Album(UnicodeMixin):
             self.orig_path = join(self.dst_path, self.settings['orig_dir'])
             check_or_create_dir(self.orig_path)
 
+    def sort_medias(self, medias_sort_attr):
+        if self.medias:
+            if medias_sort_attr == 'date':
+                key = lambda s: s.date or datetime.now()
+            else:
+                key = lambda s: strxfrm(getattr(s, medias_sort_attr))
+
+            self.medias.sort(key=key,
+                             reverse=self.settings['medias_sort_reverse'])
+
+        signals.medias_sorted.send(self)
+
     @property
     def images(self):
         """List of images (:class:`~sigal.gallery.Image`)."""
@@ -399,13 +423,15 @@ class Album(UnicodeMixin):
         if zip_gallery and len(self) > 0:
             archive_path = join(self.dst_path, zip_gallery)
             archive = zipfile.ZipFile(archive_path, 'w')
+            attr = ('src_path' if self.settings['zip_media_format'] == 'orig'
+                    else 'dst_path')
 
-            if self.settings['zip_media_format'] == 'orig':
-                for p in self:
-                    archive.write(p.src_path, os.path.split(p.src_path)[1])
-            else:
-                for p in self:
-                    archive.write(p.dst_path, os.path.split(p.dst_path)[1])
+            for p in self:
+                path = getattr(p, attr)
+                try:
+                    archive.write(path, os.path.split(path)[1])
+                except OSError as e:
+                    self.logger.warn('Failed to add %s to the ZIP: %s', p, e)
 
             archive.close()
             self.logger.debug('Created ZIP archive %s', archive_path)
@@ -419,8 +445,7 @@ class Gallery(object):
     def __init__(self, settings, ncpu=None):
         self.settings = settings
         self.logger = logging.getLogger(__name__)
-        self.stats = {'image': 0, 'image_skipped': 0,
-                      'video': 0, 'video_skipped': 0}
+        self.stats = defaultdict(int)
         self.init_pool(ncpu)
         check_or_create_dir(settings['destination'])
 
@@ -431,8 +456,16 @@ class Gallery(object):
         ignore_dirs = settings['ignore_directories']
         ignore_files = settings['ignore_files']
 
+        progressChars = cycle(["/", "-", "\\", "|"])
+        if self.logger.getEffectiveLevel() >= logging.WARNING:
+            self.progressbar_target = None
+        else:
+            self.progressbar_target = Devnull()
+
         for path, dirs, files in os.walk(src_path, followlinks=True,
                                          topdown=False):
+            if self.logger.getEffectiveLevel() >= logging.WARNING:
+                print("\rCollecting albums " + next(progressChars), end="")
             relpath = os.path.relpath(path, src_path)
 
             # Test if the directory match the ignore_dirs settings
@@ -467,6 +500,11 @@ class Gallery(object):
                 album.create_output_directories()
                 albums[relpath] = album
 
+        with progressbar(albums.values(), label="Sorting media",
+                         file=self.progressbar_target) as progress_albums:
+            for album in progress_albums:
+                album.sort_medias(settings['medias_sort_attr'])
+
         self.logger.debug('Albums:\n%r', albums.values())
         signals.gallery_initialized.send(self)
 
@@ -512,36 +550,55 @@ class Gallery(object):
             self.logger.warning("No albums found.")
             return
 
-        log_func = (partial(print, colored('->', BLUE)) if sys.stdout.isatty()
-                    else self.logger.warn)
-
-        # loop on directories in reversed order, to process subdirectories
-        # before their parent
-        media_list = []
-        processor = media_list.append if self.pool else process_file
+        def log_func(x):
+            # 63 is the total length of progressbar, label, percentage, etc
+            available_length = get_terminal_size()[0] - 64
+            if x and available_length > 10:
+                return text_type(x.name)[:available_length].encode('utf-8')
+            else:
+                return ""
 
         try:
-            for album in self.albums.values():
-                if len(album) > 0:
-                    log_func(str(album))
-                    for files in self.process_dir(album, force=force):
-                        processor(files)
-                else:
-                    self.logger.info('Album %r is empty', album)
+            with progressbar(self.albums.values(), label="Collecting files",
+                             item_show_func=log_func, show_eta=False,
+                             file=self.progressbar_target) as albums:
+                media_list = [f for album in albums
+                              for f in self.process_dir(album, force=force)]
         except KeyboardInterrupt:
             sys.exit('Interrupted')
 
+        bar_opt = {'label': "Processing files",
+                   'show_pos': True,
+                   'file': self.progressbar_target}
+
         if self.pool:
+            failed_files = []
             try:
-                # map_async is needed to handle KeyboardInterrupt correctly
-                self.pool.map_async(worker, media_list).get(9999)
+                with progressbar(length=len(media_list), **bar_opt) as bar:
+                    for res in self.pool.imap_unordered(worker, media_list):
+                        if res:
+                            failed_files.append(res)
+                        next(bar)
                 self.pool.close()
                 self.pool.join()
             except KeyboardInterrupt:
                 self.pool.terminate()
                 sys.exit('Interrupted')
+            except pickle.PicklingError:
+                self.logger.critical(
+                    "Failed to process files with the multiprocessing feature."
+                    " This can be caused by some module import or object "
+                    "defined in the settings file, which can't be serialized.",
+                    exc_info=True)
+                sys.exit('Abort')
 
+            if failed_files:
+                self.remove_files(failed_files)
             print('')
+        else:
+            with progressbar(media_list, **bar_opt) as medias:
+                for media_item in medias:
+                    process_file(media_item)
 
         if self.settings['write_html']:
             writer = Writer(self.settings, index_title=self.title)
@@ -550,39 +607,42 @@ class Gallery(object):
 
         signals.gallery_build.send(self)
 
+    def remove_files(self, files):
+        self.logger.error('Some files have failed to be processed:')
+        for path, filename in files:
+            self.logger.error('  - %s/%s', path, filename)
+            album = self.albums[path]
+            for f in album.medias:
+                if f.filename == filename:
+                    self.stats[f.type + '_failed'] += 1
+                    album.medias.remove(f)
+                    break
+        self.logger.error('You can run sigal in verbose (--verbose) or debug '
+                          '(--debug) mode to get more details.')
+
     def process_dir(self, album, force=False):
         """Process a list of images in a directory."""
-
         for f in album:
             if isfile(f.dst_path) and not force:
                 self.logger.info("%s exists - skipping", f.filename)
                 self.stats[f.type + '_skipped'] += 1
             else:
-                if self.settings['keep_orig']:
-                    copy(f.src_path, join(album.orig_path, f.filename),
-                         symlink=self.settings['orig_link'])
-
                 self.stats[f.type] += 1
-                yield f.type, f.src_path, album.dst_path, self.settings
+                yield (f.type, f.path, f.filename, f.src_path, album.dst_path,
+                       self.settings)
 
 
 def process_file(args):
-    ftype, src_path, dst_path, settings = args
-    logger = logging.getLogger(__name__)
-    logger.info('Processing %s', src_path)
-
-    if logger.getEffectiveLevel() > 20:
-        print('.', end='')
-        sys.stdout.flush()
-
-    if ftype == 'image':
-        return process_image(src_path, dst_path, settings)
-    elif ftype == 'video':
-        return process_video(src_path, dst_path, settings)
+    # args => ftype, path, filename, src_path, dst_path, settings
+    processor = process_image if args[0] == 'image' else process_video
+    ret = processor(*args[3:])
+    # If the processor return an error (ret != 0), then we return the path and
+    # filename of the failed file to the parent process.
+    return args[1:3] if ret else None
 
 
 def worker(args):
     try:
-        process_file(args)
+        return process_file(args)
     except KeyboardInterrupt:
-        return 'KeyboardException'
+        pass
